/* Target-dependent code for Juniper on MIPS processors.
   Copyright 2004 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */


#include "defs.h"
#include "gdb_string.h"
#include "gdbcore.h"
#include "mips-tdep.h"
#include "regcache.h"

/* Transform the bits comprising a 32-bit register to the right
   size for regcache_raw_supply().  This is needed when MIPS_REGSIZE is 8.  */

static void
supply_32bit_reg (int regnum, const void *addr)
{
  char buf[MAX_REGISTER_SIZE];
  store_signed_integer (buf, register_size (current_gdbarch, regnum),
                        extract_signed_integer (addr, 4));
  regcache_raw_supply (current_regcache, regnum, buf);
}

static void
mips_jnpr_core_read_registers (char *sect, unsigned size, int which,
                               CORE_ADDR addr)
{
  if (which == 0 && size == 8*38) /* Juniper mipsr7k ukern */
    {
      int reg;
      for (reg = 0; reg < 32; ++reg)
        regcache_raw_supply (current_regcache, reg, sect + 8*reg);

      regcache_raw_supply (current_regcache, PS_REGNUM, sect + 8*32);
      regcache_raw_supply (current_regcache, mips_regnum (current_gdbarch)->lo, sect + 8*33);
      regcache_raw_supply (current_regcache, mips_regnum (current_gdbarch)->hi, sect + 8*34);
      regcache_raw_supply (current_regcache, mips_regnum (current_gdbarch)->badvaddr, sect + 8*35);
      regcache_raw_supply (current_regcache, mips_regnum (current_gdbarch)->cause, sect + 8*36);
      regcache_raw_supply (current_regcache, mips_regnum (current_gdbarch)->pc, sect + 8*37);
    }

  else if (which == 0 && size == 4*40) /* Juniper mips freebsd */
    {
      int reg;
      for (reg = 0; reg < 32; ++reg)
        supply_32bit_reg (reg, sect + 4*reg);

      supply_32bit_reg (PS_REGNUM, sect + 4*32);
      supply_32bit_reg (mips_regnum (current_gdbarch)->lo, sect + 4*33);
      supply_32bit_reg (mips_regnum (current_gdbarch)->hi, sect + 4*34);
      supply_32bit_reg (mips_regnum (current_gdbarch)->badvaddr, sect + 4*35);
      supply_32bit_reg (mips_regnum (current_gdbarch)->cause, sect + 4*36);
      supply_32bit_reg (mips_regnum (current_gdbarch)->pc, sect + 4*37);
      /* FIXME: two more registers? */
    }

  else if (which == 2 && size == 4*33) /* Juniper mips freebsd */
    {
      int reg;
      for (reg = 0; reg < 32; ++reg)
        supply_32bit_reg (FP0_REGNUM + reg, sect + 4*reg);

      supply_32bit_reg (mips_regnum (current_gdbarch)->fp_control_status,
                        sect + 4*32);
    }
  else
    warning ("unrecognized regset %d in core file", which);
}

/* Register that we are able to handle jnpr core file formats. */

static struct core_fns mips_jnpr_core_fns =
  {
    bfd_target_elf_flavour,     /* core_flavour */
    default_check_format,       /* check_format */
    default_core_sniffer,       /* core_sniffer */
    mips_jnpr_core_read_registers, /* core_read_registers */
    NULL                        /* next */
  };

void
_initialize_mips_jnpr ()
{
  deprecated_add_core_fns (&mips_jnpr_core_fns);
}
