/* Target-dependent code for Juniper on PPC processors.
   Copyright 2004 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */


#include "defs.h"
#include "gdb_string.h"
#include "gdbarch.h"
#include "gdbcore.h"
#include "ppc-tdep.h"
#include "regcache.h"
#include "bfd.h"
#include "elf-bfd.h"

static void
ppc_jnpr_core_read_registers (char *sect, unsigned size, int which,
                              CORE_ADDR addr)
{
  struct gdbarch_tdep *tdep = gdbarch_tdep (current_gdbarch);

  if (which == 0 && size == 4*71) /* Juniper ppc603e ukern */
    {
      int reg;
      for (reg = 0; reg < 32; ++reg)
        regcache_raw_supply (current_regcache, reg, sect + 4*reg);

      for (reg = 0; reg < 32; ++reg)
        regcache_raw_supply (current_regcache, FP0_REGNUM + reg, sect + 4*32 + 4*reg);

      regcache_raw_supply (current_regcache, PC_REGNUM, sect + 4*64);
      regcache_raw_supply (current_regcache, tdep->ppc_ps_regnum, sect + 4*65);
      regcache_raw_supply (current_regcache, tdep->ppc_cr_regnum, sect + 4*66);
      regcache_raw_supply (current_regcache, tdep->ppc_lr_regnum, sect + 4*67);
      regcache_raw_supply (current_regcache, tdep->ppc_ctr_regnum, sect + 4*68);
      regcache_raw_supply (current_regcache, tdep->ppc_xer_regnum, sect + 4*69);
      if (tdep->ppc_mq_regnum != -1)
        regcache_raw_supply (current_regcache, tdep->ppc_mq_regnum, sect + 4*70);
    }
  else
    warning ("unrecognized regset %d in core file", which);
}

static int
jnpr_core_sniffer (struct core_fns *our_fns, bfd *abfd)
{
  int result;

  result = (bfd_get_flavour (abfd) == our_fns -> core_flavour &&
	    (elf_elfheader(abfd)->e_ident[EI_OSABI] == ELFOSABI_NONE || 
	     elf_elfheader(abfd)->e_ident[EI_OSABI] == ELFOSABI_STANDALONE));
  return (result);
}

static struct core_fns ppc_jnpr_core_fns =
  {
    bfd_target_elf_flavour,     /* core_flavour */
    default_check_format,       /* check_format */
    jnpr_core_sniffer,       /* core_sniffer */
    ppc_jnpr_core_read_registers, /* core_read_registers */
    NULL                        /* next */
  };

void
_initialize_ppc_jnpr ()
{
  deprecated_add_core_fns (&ppc_jnpr_core_fns);
}
